'''
This module acts as the most basic interface between python and mountainlab
processors. Defining a processors inputs, outputs, and params here allows
processors written in any language (as long as they follow mountainlab
conventions) to be made into a callable python function with passable params.

These functions serve as the building blocks for subpipelines, defined in
`ms4_franklab_pyplines`

In general, each function corresponds to one processor, except in the case
where multiple processors always function together

Note that NO default params are set here. This is to prevent the use of any
default values unknowingly. Default values should be provided in the pyplines
script

AKGillespie based on code from JMagland
'''

import json
import os
from collections import defaultdict

from mountainlab_pytools import mlproc as mlp
from mountainlab_pytools import mdaio


def read_dataset_params(dataset_dir):
    '''

    Parameters
    ----------
    dataset_dir : str

    Returns
    -------
    parameter_file_text : str

    '''
    params_fname = mlp.realizeFile(os.path.join(dataset_dir, 'params.json'))
    if not os.path.exists(params_fname):
        raise Exception(
            'Dataset parameter file does not exist: ' + params_fname)
    with open(params_fname) as f:
        return json.load(f)


def bandpass_filter(timeseries, timeseries_out, samplerate, freq_min,
                    freq_max, opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to .mda file
    timeseries_out : str
        Filepath to filt.mda.prv file
    samplerate : float
    freq_min : float
    freq_max : float
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'ephys.bandpass_filter',
        {
            'timeseries': timeseries
        }, {
            'timeseries_out': timeseries_out
        },
        {
            'samplerate': samplerate,
            'freq_min': freq_min,
            'freq_max': freq_max
        },
        opts
    )


def whiten(timeseries, timeseries_out, opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to 'filt.mda.prv' file
    timeseries_out : str
        Filepath to 'pre.mda.prv' file
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'ephys.whiten',
        {
            'timeseries': timeseries
        },
        {
            'timeseries_out': timeseries_out
        },
        {},
        opts
    )


def mask_out_artifacts(timeseries, timeseries_out, threshold, interval_size,
                       opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to filt.mda.prv file
    timeseries_out : str
        Filepath to filt.mda.prv file
    threshold : float
    interval_size : int
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'ms3.mask_out_artifacts',
        {
            'timeseries': timeseries
        },
        {
            'timeseries_out': timeseries_out
        },
        {
            'threshold': threshold,
            'interval_size': interval_size
        },
        opts
    )


def ms4alg(timeseries, geom, firings_out, detect_sign, adjacency_radius,
           detect_threshold, opts=None):
    '''
    Parameters
    ----------
    timeseries : str
        Filepath to `pre.mda.prv` file
    geom : list
    firings_out : str
        Filepath to `firings_raw.mda`
    detect_sign : int
    detect_threshold : float
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    pp = {'detect_sign': detect_sign,
          'adjacency_radius': adjacency_radius,
          'detect_threshold': detect_threshold,
          }
    return mlp.runProcess(
        'ms4alg.sort',
        {
            'timeseries': timeseries,
            'geom': geom
        },
        {
            'firings_out': firings_out
        },
        pp,
        opts
    )


def compute_cluster_metrics(timeseries, firings, metrics_out, samplerate,
                            opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to `pre.mda.prv` file
    firings : str
        Filepath to `firings_raw.mda` file
    metrics_out : str
        Filepath to `metrics_raw.json` file
    samplerate : float
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    metrics1 = mlp.runProcess(
        'ms3.cluster_metrics',
        {
            'timeseries': timeseries,
            'firings': firings
        },
        {
            'cluster_metrics_out': True
        },
        {
            'samplerate': samplerate
        },
        opts
    )['cluster_metrics_out']

    metrics2 = mlp.runProcess(
        'ms3.isolation_metrics',
        {
            'timeseries': timeseries,
            'firings': firings
        },
        {
            'metrics_out': True
        },
        {
            'compute_bursting_parents': 'true'
        },
        opts
    )['metrics_out']

    return mlp.runProcess(
        'ms3.combine_cluster_metrics',
        {
            'metrics_list': [metrics1, metrics2]
        },
        {
            'metrics_out': metrics_out  # True #metrics_out
        },
        {},
        opts
    )

# UNTESTED?UNUSED BY AKG


def automated_curation(firings, cluster_metrics, firings_out, opts=None):
    '''

    Parameters
    ----------
    firings : str
        Filepath
    cluster_metrics : str
        Filepath
    firings_out : str
        Filepath
    '''
    if opts is None:
        opts = {}
    label_map = mlp.runProcess(
        'ms4alg.create_label_map',
        {
            'metrics': cluster_metrics
        },
        {
            'label_map_out': True
        },
        {},
        opts
    )['label_map_out']
    return mlp.runProcess(
        'ms4alg.apply_label_map',
        {
            'label_map': label_map,
            'firings': firings
        },
        {
            'firings_out': firings_out
        },
        {},
        opts
    )


def pyms_merge_burst_parents(firings, metrics, firings_out, opts=None):
    '''

    Parameters
    ----------
    firings : str
        Filepath to `firings_raw.mda`
    metrics : str
        Filepath to `metrics_raw`
    firings_out : str
        Filepath to `firings_burst_merged.mda`
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'pyms.merge_burst_parents',
        {
            'firings': firings,
            'metrics': metrics
        },
        {
            'firings_out': firings_out
        },
        opts
    )


def tagged_curation(cluster_metrics, metrics_tagged,
                    firing_rate_thresh=0.01, isolation_thresh=0.95,
                    noise_overlap_thresh=0.03, peak_snr_thresh=1.5,
                    mv2file='', opts=None):
    '''

    Parameters
    ----------
    cluster_metrics : str
        Filepath
    metrics_tagged : str
        Filepath
    firing_rate_thresh : float, optional
    isolation_thresh : float, optional
    noise_overlap_thresh : float, optional
    peak_snr_thresh : float, optional
    mv2file : str, optional
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'pyms.add_curation_tags',
        {
            'metrics': cluster_metrics
        },
        {
            'metrics_tagged': metrics_tagged
        },
        {
            'firing_rate_thresh': firing_rate_thresh,
            'isolation_thresh': isolation_thresh,
            'noise_overlap_thresh': noise_overlap_thresh,
            'peak_snr_thresh': peak_snr_thresh,
            'mv2file': mv2file,
        },
        opts
    )


def get_mda_list(date, ntrode, data_location):
    '''

    Parameters
    ----------
    date : int
    ntrode : int
    data_location : str

    Returns
    -------
    mda_list : list of str

    '''
    date = str(date)
    mda_src_dict = defaultdict(dict)

    for epdirmda in os.listdir(os.path.join(data_location, date)):
        if '.mda' in epdirmda:
            # for each nt.mda file
            for eptetmda in os.listdir(
                    os.path.join(data_location, date, epdirmda)):
                if '.nt' in eptetmda:
                    ep = eptetmda.split('_')[2].split('.')[0]
                    ntr = eptetmda.split('_')[-1].split('.')[1]
                    mda_src_dict[ntr][ep] = os.path.join(
                        data_location, date, epdirmda, eptetmda)
    mda_list = list(mda_src_dict[f'nt{ntrode}'].values())
    mda_list.sort()

    return mda_list


def get_epoch_offsets(dataset_dir, opts=None):
    '''
    Parameters
    ----------
    dataset_dir : str
    opts : None or dict, optional

    Returns
    -------
    sample_offsets : ???
    total_samples : ???

    '''
    if opts is None:
        opts = {}
    if 'mda_list' in opts:
        # initialize with 0 (first start time)
        lengths = [0]

        for idx in range(len(opts['mda_list'])):
            ep_path = opts['mda_list'][idx]
            ep_mda = mdaio.DiskReadMda(ep_path)
            # get length of the mda (N dimension)
            samplength = ep_mda.N2()
            # add to prior sum and append
            lengths.append(samplength + lengths[(idx)])

    else:

        prv_list = os.path.join(dataset_dir, 'raw.mda.prv')

        with open(prv_list, 'r') as f:
            ep_files = json.load(f)

        # initialize with 0 (first start time)
        lengths = [0]

        for idx in range(len(ep_files['files'])):
            ep_path = ep_files['files'][idx]['prv']['original_path']
            ep_mda = mdaio.DiskReadMda(ep_path)
            # get length of the mda (N dimension)
            samplength = ep_mda.N2()
            # add to prior sum and append
            lengths.append(samplength + lengths[(idx)])

    # first entries (incl 0) are starttimes; last is total time
    total_samples = lengths[-1]
    sample_offsets = lengths[0:-1]

    return sample_offsets, total_samples


def pyms_extract_segment(timeseries, timeseries_out, t1, t2, opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath
    timeseries_out : str
        Filepath
    t1 : float
    t2 : float
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}

    return mlp.runProcess(
        'pyms.extract_timeseries',
        {
            'timeseries': timeseries
        },
        {
            'timeseries_out': timeseries_out
        },
        {
            't1': t1,
            't2': t2
        },
        opts
    )


def pyms_anneal_segs(timeseries_list, firings_list, firings_out,
                     dmatrix_out, k1_dmatrix_out, k2_dmatrix_out,
                     dmatrix_templates_out, time_offsets, opts=None):
    '''

    Parameters
    ----------
    timeseries_list : list of str
        Filepaths
    firings_list : list of str
        Filepaths
    firings_out : list of str
        Filepaths
    dmatrix_out : list
    k1_dmatrix_out : list
    k2_dmatrix_out : list
    time_offsets : str
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}

    return mlp.runProcess(
        'pyms.anneal_segments',
        {
            'timeseries_list': timeseries_list,
            'firings_list': firings_list
        },
        {
            'firings_out': firings_out,
            'dmatrix_out': dmatrix_out,
            'k1_dmatrix_out': k1_dmatrix_out,
            'k2_dmatrix_out': k2_dmatrix_out,
            'dmatrix_templates_out': dmatrix_templates_out
        },
        {
            'time_offsets': time_offsets
        },
        opts
    )


def clear_seg_files(timeseries_list, firings_list, opts=None):
    '''

    Parameters
    ----------
    timeseries_list : list of str
        Filepaths
    firings_list : list of str
    opts : None or dict

    '''
    if opts is None:
        opts = {}
    for file in timeseries_list:
        os.remove(file)

    for file in firings_list:
        os.remove(file)


def pyms_extract_clips(timeseries, firings, clips_out, opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to `pre.mda.prv` file
    firings : str
        Filepath to `firings_raw.mda` file
    clips_out : str
        Filepath to `clips.mda` file
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'pyms.extract_clips',
        {
            'timeseries': timeseries,
            'firings': firings
        },
        {
            'clips_out': clips_out
        },
        opts
    )


def pyms_extract_marks(timeseries, firings, marks_out, markstimes_out,
                       opts=None):
    '''

    Parameters
    ----------
    timeseries : str
        Filepath to `pre.mda.prv`
    firings : str
        Filepath to `firings_raw.mda`
    marks_out : str
        Filepath to `marks.mda`
    markstimes_out : str
        Filepath
    opts : None or dict, optional
    '''
    if opts is None:
        opts = {}
    return mlp.runProcess(
        'pyms.extract_marks',
        {
            'timeseries': timeseries,
            'firings': firings
        },
        {
            'marks_out': marks_out,
            'markstimes_out': markstimes_out
        },
        opts
    )


def synthesize_sample_dataset(dataset_dir, samplerate=30000, duration=600,
                              num_channels=4, opts=None):
    '''

    Parameters
    ----------
    dataset_dir : str
    samplerate : float
    duration : float
    num_channels : int
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    if not os.path.exists(dataset_dir):
        os.mkdir(dataset_dir)
    M = num_channels
    mlp.runProcess(
        'ephys.synthesize_random_waveforms',
        {},
        {
            'geometry_out': os.path.join(dataset_dir, 'geom.csv'),
            'waveforms_out': os.path.join(dataset_dir, 'waveforms_true.mda')
        },
        {
            'upsamplefac': 13,
            'M': M,
            'average_peak_amplitude': 100
        },
        opts
    )
    mlp.runProcess(
        'ephys.synthesize_random_firings',
        {},
        {
            'firings_out': os.path.join(dataset_dir, 'firings_true.mda')
        },
        {
            'duration': duration
        },
        opts
    )
    mlp.runProcess(
        'ephys.synthesize_timeseries',
        {
            'firings': os.path.join(dataset_dir, 'firings_true.mda'),
            'waveforms': os.path.join(dataset_dir, 'waveforms_true.mda')
        },
        {
            'timeseries_out': os.path.join(dataset_dir, 'raw.mda.prv')
        }, {
            'duration': duration,
            'waveform_upsamplefac': 13,
            'noise_level': 10
        },
        opts
    )
    params = {
        'samplerate': samplerate,
        'spike_sign': 1
    }
    with open(os.path.join(dataset_dir, 'params.json'), 'w') as outfile:
        json.dump(params, outfile, indent=4)



#!/usr/bin/env python3
from setuptools import find_packages, setup

INSTALL_REQUIRES = []
TESTS_REQUIRE = []

setup(
    name='franklab_mountainsort',
    version='0.1.0.dev0',
    license='',
    description=(''),
    author='',
    author_email='',
    url='',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
)



'''
This module calls the helper functions defined in p2p (`proc2py`) that in turn,
call Mountain Sort processors. This should be a collection of common processing
steps that are standard across the lab, although parameters can be changed
flexibly. Default parameters are defined in the arguments of each pypline, but
can be overwritten by the user.

These pyplines should be called by a python batch script, which will manage
running the steps on particular animal, days, and ntrodes.

AKGillespie based on code from JMagland
Demetris Roumis

# WARNING: Before anything else, must concat all eps together becuase ms4 no
           longer handles the prv list of mdas
'''
import json
import logging
import math
import os
import subprocess

from franklab_msdrift.p_anneal_segments import \
    anneal_segments as pyms_anneal_segs

from franklab_mountainsort.ms4_franklab_proc2py import (bandpass_filter,
                                                        clear_seg_files,
                                                        compute_cluster_metrics,
                                                        get_epoch_offsets,
                                                        get_mda_list,
                                                        mask_out_artifacts,
                                                        ms4alg,
                                                        pyms_extract_clips,
                                                        pyms_extract_segment,
                                                        read_dataset_params,
                                                        tagged_curation,
                                                        whiten)
from franklab_mstaggedcuration.p_add_curation_tags import \
    add_curation_tags as pyms_add_curation_tags
from franklab_mstaggedcuration.p_merge_burst_parents import \
    merge_burst_parents as pyms_merge_burst_parents


def concat_epochs(dataset_dir, mda_list=None, opts=None, mda_opts=None):
    '''Concatenate all epochs in a day and save as raw.mda.

    Saves the raw.mda to the output dir, which serves as src for subsequent
    steps.

    Runs 'ms3.concat_timeseries' using either:
       1: mda_list provided
       2: mda_list empty and date, ntrode specified in opts
       3: string path to prv file containing entries for mda files

    Parameters
    ----------
    dataset_dir : str
    mda_list : None or list, optional
    opts : None or dict, optional
    mda_opts : None or dict, optional

    Notes
    -----
    Format for input and output of ms3.concat_timeseries is:
        'timeseries_list:{path} timeseries_list:{path} ...'
    There cannot be a space between the colon and the path.

    '''

    if mda_list is None:
        mda_list = []
    if opts is None:
        opts = {}
    if mda_opts is None:
        mda_opts = {}

    strstart = []
    if isinstance(mda_list, list) and len(mda_list) > 0:
        logging.info('Using provided list of mda files')
        for entry in mda_list:
            strstart.append(f'timeseries_list:{entry}')
    has_opts_keys = (
        {'anim', 'date', 'ntrode', 'data_location'}.issubset(mda_opts))
    if len(mda_list) == 0 and has_opts_keys:
        logging.info(
            f'Finding list of mda file from mda directories of '
            f'date: {mda_opts["date"]}, ntrode: {mda_opts["ntrode"]}')
        mda_list = get_mda_list(
            mda_opts['date'], mda_opts['ntrode'], mda_opts['data_location'])
        for entry in mda_list:
            strstart.append(f'timeseries_list:{entry}')

    if isinstance(mda_list, str):
        logging.info('Using mda files listed in prv file')
        with open(mda_list) as f:
            mdalist = json.load(f)
        for entries in mdalist['files']:
            prv_path = entries['prv']['original_path']
            strstart.append(f'timeseries_list:{prv_path}')

    joined = ' '.join(strstart)
    outpath = os.path.join(f'timeseries_out:{dataset_dir}', 'raw.mda')
    subprocess.run(['ml-run-process', 'ms3.concat_timeseries',
                    '--inputs', joined, '--outputs', outpath], check=True)


def filt_mask_whiten(dataset_dir, output_dir, freq_min=300, freq_max=6000,
                     mask_artifacts=True, opts=None):
    '''

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    freq_min : float, optional
    freq_max : float, optional
    mask_artifacts : bool, optional
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    # Dataset parameters
    ds_params = read_dataset_params(dataset_dir)

    # Bandpass filter
    bandpass_filter(
        timeseries=os.path.join(dataset_dir, 'raw.mda'),
        timeseries_out=os.path.join(output_dir, 'filt.mda.prv'),
        samplerate=ds_params['samplerate'],
        freq_min=freq_min,
        freq_max=freq_max,
        opts=opts
    )
    # Mask out artifacts
    if mask_artifacts:
        mask_out_artifacts(
            timeseries=os.path.join(output_dir, 'filt.mda.prv'),
            timeseries_out=os.path.join(output_dir, 'filt.mda.prv'),
            threshold=5,
            interval_size=2000,
            opts=opts
        )
    # Whiten
    whiten(
        timeseries=os.path.join(output_dir, 'filt.mda.prv'),
        timeseries_out=os.path.join(output_dir, 'pre.mda.prv'),
        opts=opts
    )


# full = sort the entire file as one mda
def ms4_sort_full(dataset_dir, output_dir, geom=None, adjacency_radius=-1,
                  detect_threshold=3, detect_sign=False, opts=None):
    '''
    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    geom : None or list, optional
    adjacency_radius : float, optional
    detect_threshold : float, optional
    detect_sign : bool, optional
    opt : dict or None, optional

    '''
    if geom is None:
        geom = []
    if opts is None:
        opts = {}
    # Fetch dataset parameters
    ds_params = read_dataset_params(dataset_dir)

    ms4alg(
        timeseries=os.path.join(output_dir, 'pre.mda.prv'),
        geom=geom,
        firings_out=os.path.join(output_dir, 'firings_raw.mda'),
        adjacency_radius=adjacency_radius,
        detect_sign=int(detect_sign),
        detect_threshold=detect_threshold,
        opts=opts
    )

    # Compute cluster metrics
    compute_cluster_metrics(
        timeseries=os.path.join(output_dir, 'pre.mda.prv'),
        firings=os.path.join(output_dir, 'firings_raw.mda'),
        metrics_out=os.path.join(output_dir, 'metrics_raw.json'),
        samplerate=ds_params['samplerate'],
        opts=opts
    )


def ms4_sort_on_segs(dataset_dir, output_dir, geom=None,
                     adjacency_radius=-1, detect_threshold=3.0,
                     detect_sign=False, rm_segment_intermediates=True,
                     opts=None, mda_opts=None):
    '''Sort by timesegments, then join any matching clusters

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    geom : None or list, optional
    adjacency_radius : float, optional
    detect_threshold : float, optional
    detect_sign : bool, optional
    rm_segment_intermediates : bool, optional
    opt : dict or None, optional
    mda_opt : dict or None, optional

    '''
    if geom is None:
        geom = []
    if opts is None:
        opts = {}
    if mda_opts is None:
        mda_opts = {}

    # Fetch dataset parameters
    ds_params = read_dataset_params(dataset_dir)
    has_keys = {'anim', 'date', 'ntrode', 'data_location'}.issubset(mda_opts)

    if has_keys:
        logging.info(
            'Finding list of mda file from mda directories of '
            f'date:{mda_opts["date"]}, ntrode:{mda_opts["ntrode"]}')
        mda_list = get_mda_list(
            mda_opts['date'], mda_opts['ntrode'], mda_opts['data_location'])
        # calculate time_offsets and total_duration
        sample_offsets, total_samples = get_epoch_offsets(
            dataset_dir=dataset_dir, opts={'mda_list': mda_list})

    else:
        # calculate time_offsets and total_duration
        sample_offsets, total_samples = get_epoch_offsets(
            dataset_dir=dataset_dir)

    # break up preprocesed data into segments and sort each
    firings_list = []
    timeseries_list = []
    for segind in range(len(sample_offsets)):
        t1 = math.floor(sample_offsets[segind])
        if segind == len(sample_offsets) - 1:
            t2 = total_samples - 1
        else:
            t2 = math.floor(sample_offsets[segind + 1]) - 1

        t1_min = t1 / ds_params['samplerate'] / 60
        t2_min = t2 / ds_params['samplerate'] / 60
        logging.info(f'Segment {segind + 1}: t1={t1}, t2={t2}, '
                     f't1_min={t1_min:.3f}, t2_min={t2_min:.3f}')

        pre_outpath = os.path.join(dataset_dir, f'pri-{segind + 1}.mda')
        pyms_extract_segment(
            timeseries=os.path.join(output_dir, 'pre.mda.prv'),
            timeseries_out=pre_outpath,
            t1=t1,
            t2=t2,
            opts=opts)

        firings_outpath = os.path.join(
            dataset_dir, f'firings-{segind + 1}.mda')
        ms4alg(
            timeseries=pre_outpath,
            firings_out=firings_outpath,
            geom=geom,
            detect_sign=int(detect_sign),
            adjacency_radius=adjacency_radius,
            detect_threshold=detect_threshold,
            opts=opts)

        firings_list.append(firings_outpath)
        timeseries_list.append(pre_outpath)

    firings_out_final = os.path.join(output_dir, 'firings_raw.mda')
    # sample_offsets have to be converted into a string to be properly passed
    # into the processor
    str_sample_offsets = ','.join(map(str, sample_offsets))
    logging.info(str_sample_offsets)

    pyms_anneal_segs(
        timeseries_list=timeseries_list,
        firings_list=firings_list,
        firings_out=firings_out_final,
        dmatrix_out=[],
        k1_dmatrix_out=[],
        k2_dmatrix_out=[],
        dmatrix_templates_out=[],
        time_offsets=str_sample_offsets
    )

    # clear the temp pre and firings files if specified
    if rm_segment_intermediates:
        clear_seg_files(
            timeseries_list=timeseries_list,
            firings_list=firings_list
        )

    # Compute cluster metrics
    compute_cluster_metrics(
        timeseries=os.path.join(output_dir, 'pre.mda.prv'),
        firings=os.path.join(output_dir, 'firings_raw.mda'),
        metrics_out=os.path.join(output_dir, 'metrics_raw.json'),
        samplerate=ds_params['samplerate'],
        opts=opts
    )


def merge_burst_parents(dataset_dir, output_dir):
    '''

    Parameters
    ----------
    dataset_dir : str
    output_dir : str

    '''
    pyms_merge_burst_parents(
        firings=os.path.join(output_dir, 'firings_raw.mda'),
        metrics=os.path.join(output_dir, 'metrics_raw.json'),
        firings_out=os.path.join(output_dir, 'firings_burst_merged.mda'))

    ds_params = read_dataset_params(dataset_dir)
    # Compute cluster metrics
    compute_cluster_metrics(
        timeseries=os.path.join(output_dir, 'pre.mda.prv'),
        firings=os.path.join(output_dir, 'firings_burst_merged.mda'),
        metrics_out=os.path.join(output_dir, 'metrics_merged.json'),
        samplerate=ds_params['samplerate'])


def add_curation_tags(dataset_dir, output_dir, firing_rate_thresh=0.01,
                      isolation_thresh=0.95, noise_overlap_thresh=0.03,
                      peak_snr_thresh=1.5, metrics_input='metrics_raw.json',
                      metrics_output='metrics_tagged.json'):
    '''

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    firing_rate_thresh : float, optional
    isolation_thresh : float, optional
    noise_overlap_thresh : float, optional
    peak_snr_thresh : float, optional
    metrics_input : str, optional
    metrics_output : str, optional
    opts : None or dict, optional

    Notes
    -----
    This is split out and not included after metrics calculation
    because of a bug in ms3.combine_cluster_metrics - doesn't work if anything
    follows it.

    '''
    pyms_add_curation_tags(
        metrics=os.path.join(dataset_dir, metrics_input),
        metrics_tagged=os.path.join(output_dir, metrics_output),
        firing_rate_thresh=firing_rate_thresh,
        isolation_thresh=isolation_thresh,
        noise_overlap_thresh=noise_overlap_thresh,
        peak_snr_thresh=peak_snr_thresh, mv2file='')


def recalc_metrics(dataset_dir, output_dir, firings_in='',
                   metrics_to_update='', firing_rate_thresh=0.01,
                   isolation_thresh=0.95, noise_overlap_thresh=0.03,
                   peak_snr_thresh=1.5, mv2_file='', opts=None):
    '''post-merge, should recalculate metrics and update tags (both tags based
    on thresholds and any manually added ones, stored in the mv2)

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    firings_in : str, optional
    metrics_to_update : str, optional
    firing_rate_thresh : float, optional
    isolation_thresh : float, optional
    noise_overlap_thresh : float, optional
    peak_snr_thresh : float, optional
    mv2_file : str, optional
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    # untested!
    if not firings_in:
        firings_in = 'firings_processed.json'
    if not metrics_to_update:
        metrics_to_update = 'metrics_tagged.json'

    ds_params = read_dataset_params(dataset_dir)

    compute_cluster_metrics(
        timeseries=os.path.join(output_dir, 'pre.mda.prv'),
        firings=os.path.join(output_dir, firings_in),
        metrics_to_update=os.path.join(output_dir, metrics_to_update),
        samplerate=ds_params['samplerate'])

    tagged_curation(
        cluster_metrics=os.path.join(dataset_dir, metrics_to_update),
        metrics_tagged=os.path.join(output_dir, metrics_to_update),
        firing_rate_thresh=firing_rate_thresh,
        isolation_thresh=isolation_thresh,
        noise_overlap_thresh=noise_overlap_thresh,
        peak_snr_thresh=peak_snr_thresh,
        mv2file=mv2_file,
        opts=opts
    )


def extract_clips(dataset_dir, output_dir, clip_size=100, opts=None):
    '''

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    clip_size : float, optional
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    opts['clip_size'] = clip_size

    pyms_extract_clips(
        timeseries=os.path.join(dataset_dir, 'pre.mda.prv'),
        firings=os.path.join(dataset_dir, 'firings_raw.mda'),
        clips_out=os.path.join(output_dir, 'clips.mda'),
        opts=opts)


def extract_marks(dataset_dir, output_dir, opts=None):
    '''

    Parameters
    ----------
    dataset_dir : str
    output_dir : str
    opts : None or dict, optional

    '''
    if opts is None:
        opts = {}
    opts['clip_size'] = 1

    pyms_extract_clips(
        timeseries=os.path.join(dataset_dir, 'pre.mda.prv'),
        firings=os.path.join(dataset_dir, 'firings_raw.mda'),
        clips_out=os.path.join(output_dir, 'marks.mda'),
        opts=opts)



import errno
import glob
import json
import logging
import os
import subprocess

import franklab_mountainsort.ms4_franklab_pyplines as pyp

logging.basicConfig(level='INFO', format='%(asctime)s %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S')


def move_mda_data(source_animal_path, target_animal_path, animal, dates):
    '''Move data from preprocessing to scratch.

    Parameters
    ----------
    source_animal_path : str
    target_animal_path : str
    animal : str
    dates : list of int

    '''
    for date in dates:
        date = str(date)
        source_mda_paths = glob.glob(
            os.path.join(source_animal_path, date, '*.mda'))
        source_mda_paths.sort()
        for source_path in source_mda_paths:
            target_dir = os.path.join(target_animal_path, date)
            os.makedirs(target_dir, exist_ok=True)
            logging.info(f'\nCopying {source_path} to {target_dir}\n')
            subprocess.run(
                f'rsync -avP {source_path} {target_dir}', shell=True)


def run_spike_sorting(animal, dates, ntrodes, input_path, output_path,
                      metrics_input='metrics_merged.json',
                      metrics_output='metrics_merged_tagged.json',
                      firing_rate_thresh=0.01, isolation_thresh=0.95,
                      noise_overlap_thresh=0.03, peak_snr_thresh=1.5,
                      extract_marks=True, extract_clips=True,
                      clip_size=100, freq_min=300, freq_max=6000,
                      adjacency_radius=-1, detect_threshold=3, detect_sign=-1,
                      sampling_rate=30000):
    '''Runs mountain sort on data.

    Parameters
    ----------
    animal : str
    dates : list of str
    ntrodes : list of int
    input_path : str
    output_path : str
    metrics_input : str, optional
    metrics_output : str, optional
    firing_rate_thresh : float, optional
    isolation_thresh : float, optional
    noise_overlap_thresh : float, optional
    peak_snr_thresh : float, optional
    extract_marks : bool, optional
    extract_clips : bool, optional
    clip_size : float, optional
         The size of extract clips around each spike in samples.
    freq_min : float, optional
        The highpass or low cutoff of the filter in Hz.
    freq_max : float, optional
        The lowpass or high cutoff of the filter in Hz.
    adjacency_radius : float, optional
        The radius in µm that defines a neighborhood of channels on which to
        sort (default -1 to ignore and not require a geom.csv file, useful for
        tetrodes).
    detect_threshold : float, optional
        Spike detection threshold in standard deviations.
    detect_sign : int, optional
         The direction of spike to detect (-1 for negative, 1 for positive,
         0 for both). -1 is recommended for most recordings.
    sampling_rate : int, optional
        Number of samples per second.

    '''
    # Make .mnt symlinks if they don't exist
    date_folders = os.path.join(input_path, '*')
    if len(glob.glob(os.path.join(date_folders, '*.mnt'))) == 0:
        make_mda_ntrodeEpoch_links(input_path)

    for date in dates:
        date = str(date)
        logging.info(f'Running {animal} date: {date} ntrodes: {ntrodes}')
        mountain_mda_path = os.path.join(
            input_path, date, f'{date}_{animal}.mountain')
        mountain_out_path = os.path.join(output_path, date, 'ms4')

        for electrode_number in ntrodes:
            mountain_mda_nt_path = os.path.join(
                mountain_mda_path, f'nt{electrode_number}')
            mountain_out_nt_path = os.path.join(
                mountain_out_path, f'nt{electrode_number}')
            os.makedirs(mountain_out_nt_path, exist_ok=True)
            mda_opts = {'anim': animal,
                        'date': date,
                        'ntrode': electrode_number,
                        'data_location': input_path}
            raw_mda = os.path.join(mountain_mda_nt_path, 'raw.mda')

            # create concatenated mda if it doesn't exist
            if not os.path.isfile(raw_mda):
                os.makedirs(mountain_mda_nt_path, exist_ok=True)
                # create params if it doesn't exist
                params_file = os.path.join(mountain_mda_nt_path, 'params.json')
                if not os.path.isfile(params_file):
                    params = {'samplerate': sampling_rate}
                    with open(params_file, 'w') as f:
                        json.dump(params, f, indent=4, sort_keys=True)
                logging.info(f'Creating concatenated epochs .mda: {raw_mda}')
                pyp.concat_epochs(dataset_dir=mountain_mda_nt_path,
                                  mda_opts=mda_opts)

            logging.info(f'####### NTRODE_INPUT: {raw_mda}')
            logging.info(f'####### NTRODE_OUTPUT: {mountain_out_nt_path}')

            pyp.filt_mask_whiten(
                dataset_dir=mountain_mda_nt_path,
                output_dir=mountain_out_nt_path,
                freq_min=freq_min,
                freq_max=freq_max)
            pyp.ms4_sort_on_segs(
                dataset_dir=mountain_mda_nt_path,
                output_dir=mountain_out_nt_path,
                adjacency_radius=adjacency_radius,
                detect_threshold=detect_threshold,
                detect_sign=detect_sign,
                mda_opts=mda_opts)
            pyp.merge_burst_parents(
                dataset_dir=mountain_mda_nt_path,
                output_dir=mountain_out_nt_path)
            pyp.add_curation_tags(
                dataset_dir=mountain_out_nt_path,
                output_dir=mountain_out_nt_path,
                metrics_input=metrics_input,
                metrics_output=metrics_output,
                firing_rate_thresh=firing_rate_thresh,
                isolation_thresh=isolation_thresh,
                noise_overlap_thresh=noise_overlap_thresh,
                peak_snr_thresh=peak_snr_thresh)

            if extract_marks:
                pyp.extract_marks(dataset_dir=mountain_out_nt_path,
                                  output_dir=mountain_out_nt_path)

            if extract_clips:
                pyp.extract_clips(dataset_dir=mountain_out_nt_path,
                                  output_dir=mountain_out_nt_path,
                                  clip_size=clip_size)


def make_mda_ntrodeEpoch_links(preprocessing_path):
    '''This makes the .mnt directory and its contents, containing symbolic
    links to the mda files for each epoch for each ntrode.

    We use this because our raw data is stored by epoch, with all tets in
    subfolders. Since we generally want to sort across all the epochs in a day,
    this tool reorganizes the data so that it is first grouped by tetrode, then
    epoch. This is all done with symlinks, so no duplicate files are generated.

    Parameters
    ----------
    preprocessing_path : str
        Folder path to animal preprocessing folder e.g.
        '.../{animal}/preprocessing'

    '''
    mda_filepaths = glob.glob(
        os.path.join(preprocessing_path, '*', '*', '*.nt*.mda'))
    for mda_filepath in mda_filepaths:
        try:
            mda_filename = os.path.basename(mda_filepath)
            date, animal, epoch, ending = mda_filename.split('_')
            _, electrode_name, _ = ending.split('.')
            dest_path = os.path.join(
                preprocessing_path, date, f'{date}_{animal}.mnt',
                f'{date}_{animal}.{electrode_name}.mnt')
            os.makedirs(dest_path, exist_ok=True)
            dest_link = os.path.join(dest_path, mda_filename)

            # to overwrite. remove ntlink if it already exists
            remove_NT_file(dest_link)

            # create directory of sym links to original mda
            os.symlink(mda_filepath, dest_link)

        except ValueError:
            pass


def remove_NT_file(ntrode_filename):
    try:
        os.remove(ntrode_filename)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred
